#include "Include/all.h"
#define million 1000000

/************************************************************************
  elapsetime.c

    Keep track of the time and memory used by the current process.
    The code was taken from that of EST by John Giraldi, and modified by
    the author.
    Different time statistics are kept as static variables of this
    module, so statistic reporting functions are also in this file.

  Change Log
  
  Date     Ini Comment 
  -------- --- -------
  02/24/92 PC  Move all time statics, and functions that utilizes those
	       variables into this module to improve modularity 
  03/09/92 PC  Add a time structure, elapsetime_s, and elapsetime_e to
               handle keeping track of nested times

  2/18/92 PeteChen
  9/20/92 X. Chen
 ************************************************************************/

char Time_trans[MAXTIMENUM][32] =
{"PARSING", "SFSIM", "RESTORE_COMPACTOR", "OMISSION_COMPACTOR", "TOTAL"};

static float runtime;
static struct Time_struct {
  float interval; /* the amount of time accumulated so far */
  float stamp;    /* the time stamp for the last elpasetime_s call */
  short flag;     /* the flag indicating starting or ending time */
} Time[MAXTIMENUM+1];
static int init;  /* set to 1 if Time structures are initialized */

#define time_interval(i) (Time[(i)].interval)
#define time_stamp(i)    (Time[(i)].stamp)
#define time_flag(i)     (Time[(i)].flag)

#ifdef SUNUX
int  getrusage();
int  getpagesize();
#endif

#ifdef HPUX
clock_t times();
long sysconf();
#endif

/*
 * Function:
 *   - keep track of the total running time
 *   - calculate the total time used since the last call
 *   - the total memory usage in megs
 *   - the total number of page faults
 * Input:
 *   file  - a file pointer to the output file
 *   note  - a string containing a note regarding this time report
 *   print - whether the time report should be output (1 yes)
 * Implicit Input:
 *   runntime - a global variable to keep track of total running time
 * Return:
 *   a float containing the number of cpu secs used since the last call
 */


float elapsetime(FILE *file, int  timelabel, int print)
{ /* elapsetime */
  int           pagesize;         

  int           who;              
  float		usertimesecs;     
  float		usertimeusecs;    
  float		systimesecs;      
  float		systimeusecs;     
  int		totalpagefaults;  
  float		totalusertime;    
  float		totalsystime;     
  float     	totaltime;	  
				  
  int           totalpage;        
  float		totalmemory;      
  float		thistime;         

  totaltime = 0.0;
  totalmemory = 0.0;
  totalpagefaults = 0;

#ifdef SUNUX
  pagesize = getpagesize();

  who      = RUSAGE_SELF;
  getrusage(who, &a_rusage);
  totalpagefaults = a_rusage.ru_majflt;
  usertimesecs    = (float) a_rusage.ru_utime.tv_sec;
  usertimeusecs   = (float) a_rusage.ru_utime.tv_usec;
  totalusertime	  = usertimesecs + usertimeusecs/million;
  systimesecs     = (float) a_rusage.ru_stime.tv_sec;
  systimeusecs    = (float) a_rusage.ru_stime.tv_usec;
  totalsystime    = systimesecs + systimeusecs/million;
  totalpage       = a_rusage.ru_maxrss;
  totalmemory     = (pagesize * totalpage) / million;

  who      = RUSAGE_CHILDREN;
  getrusage(who, &a_rusage);
  totalpagefaults = totalpagefaults + a_rusage.ru_majflt;
  usertimesecs    = (float) a_rusage.ru_utime.tv_sec;
  usertimeusecs   = (float) a_rusage.ru_utime.tv_usec;
  totalusertime   = totalusertime + usertimesecs + usertimeusecs/million;
  systimesecs     = (float) a_rusage.ru_stime.tv_sec;
  systimeusecs    = (float) a_rusage.ru_stime.tv_usec;
  totalsystime    = totalsystime + systimesecs + systimeusecs/million;
#endif

#ifdef HPUX
  times(&a_tms);
  totalusertime   = (float) (a_tms.tms_utime + a_tms.tms_cutime)/sysconf(_SC_CLK_TCK);
  totalsystime    = (float) (a_tms.tms_stime + a_tms.tms_cstime)/sysconf(_SC_CLK_TCK);
#endif

  totaltime       = totalusertime + totalsystime;
  thistime        = totaltime - runtime; 
  runtime         = totaltime;

  if (print==1)
    fprintf(file, "MEMORY(megs.) %.3f PageFaults %d\n",
	    totalmemory,totalpagefaults);

  return (thistime);
} /* elapsetime */

/*
 * Function:
 *   prints out ATPG statistics to the result file
 * Input:
 *   circuitname - a pointer to a string containing the circuit name
 * Return:
 *   none
 */
int printresults (char *circuitname, FILE *file, int sfsim_flts, int faultcnt)
{ /* printresults */
  int i;

  fprintf(file,"\n****** SPECTRALATPG Summary ******\n");
  fprintf(file,"Circuit %s", circuitname);
  fprintf (file,"   PI's: %d   Gates: %d   PO's: %d", num_pi - num_dff,
	   gatecnt - num_dff,
	   num_po - num_dff);
  if(seq_circuit == true)
    fprintf(file,"   D-F-F's: %d\n",num_dff);
  else
    fprintf(file, "\n");

  fprintf(file,"Fault Coverage:  %8.4f %c\n",
	  (float) (sfsim_flts) * 100.0 / faultcnt, '%');
  fprintf(file, "%8d %-25s\n", sfsim_flts, "FSIM Detected Faults");
  fprintf(file, "%8d %-25s\n", faultcnt, "Total Faults");

  fprintf (file, "CPU Time (sec.)\n");
  for (i=0; i < TIME_TOTAL && (i!=2 && i!=3)/*excluding the two compactor times*/; i++)
    fprintf (file, "%8.4f   %-20s\n", time_interval (i), Time_trans [i]);
  fprintf (file, "%8.4f   %-20s\n", time_interval (TIME_TOTAL) - time_interval (0)
	   - time_interval (1),
	   "OVERHEAD");
  fprintf (file,"%8.4f   %-20s\n", time_interval (TIME_TOTAL), Time_trans [TIME_TOTAL]);

  fprintf(file,"Memory Usage (KBytes):\n");
  fprintf(file,"  %6.1f   Circuit Memory\n",((float) totalcktmemory) / 1024.0);
  fprintf(file,"  %6.1f   FSIM Memory\n",((float) totalsfsimmemory) / 1024.0);
  fprintf(file,"  %6.1f   Memory Overhead\n",((float) totalothermemory) / 1024.0);
  fprintf(file,"  %6.1f   Total Memory\n",
	  ((float) (totalcktmemory + totalsfsimmemory + totalothermemory)) / 1024.0);

  return (int) time_interval(2); //total time in sec.
} /* printresults */

/*
 * Function:
 *   Print out usage times for sfsim
 * Input:
 *   file - a file pointer the file for output
 * Return:
 *   none
 */
void printsfsimtime (FILE *file)
{ /* printsfsimtime */
  fprintf(file,"%s Time = %.8f\n",Time_trans[TIME_SFSIM],
	  time_interval(TIME_SFSIM));
  fprintf(file,"TOTAL Time = %8f\n",runtime);
} /* printsfsimtime */

/*
 * Function:
 *   initialize a time structure
 * Input:
 *   i - index of the time structure
 * Return:
 *   none
 */
void time_init (int i) /* index of the Time structure to be initialized */
{ /* time_init */
  time_interval(i) = 0.0;
  time_stamp(i)    = 0.0;
  time_flag(i)     = 0;
} /* time_init */

/*
 * Function:
 *   - keep track of the total running time of a particular statistics
 *   - initialize the time structures if necessary
 * Input:
 *   file  - a file pointer to the output file
 *   note  - a string containing a note regarding this time report
 *   print - whether the time report should be output (1 yes)
 * Implicit Input:
 *   runntime - a global variable to keep track of total running time
 * Return:
 *   None
 * Note:
 *   This function and elapsetime_e are the frontend to elapsetime.
 *   They should be used in pairs with the same timelabel, similar to
 *   the way parentheses are used.  Elapsetime should not be called
 *   directly.
 */
float elapsetime_s (FILE *file, int timelabel, int print)
{ /* elapsetime_s */
  int i;

  if (init!=1) /* check whether the time structures are initialized */
    {
      for (i=0;i<MAXTIMENUM;i++)
	time_init(i);
      init = 1;
    }

  if( timelabel<0||timelabel>=MAXTIMENUM)
    {
      fprintf(stderr,"no such time label %d\n",timelabel);
      return(0);
    }
  if (time_flag(timelabel) != 0)
    {
      fprintf(stderr,"no matching elapsetime_e call for previous elapsetime_s call for %s.\n",Time_trans[timelabel]);
      return(0);
    }
  else
    {
      time_flag(timelabel) = 1;

      elapsetime(file,timelabel,print);

      time_stamp(timelabel) = runtime;
      return(0);
    }
} /* elapsetime_s */

/*
 * Function:
 *   - keep track of the total running time of a particular time
 *     staticstic 
 * Input:
 *   file  - a file pointer to the output file
 *   note  - a string containing a note regarding this time report
 *   print - whether the time report should be output (1 yes)
 * Implicit Input:
 *   runntime - a global variable to keep track of total running time
 * Return:
 *   a float containing the number of cpu secs used during this
 *   elapsetime_s and elpasetime_e pair 
 * Note:
 *   This function and elapsetime_s are the frontend to elapsetime.
 *   They should be used in pairs with the same timelabel, similar to
 *   the way parentheses are used.  Elapsetime should not be called
 *   directly.
 */
float elapsetime_e(FILE *file, int timelabel, int print)
{ /* elapsetime_e */
  float thistime;
  int i;

  if (init!=1) /* check whether the time structures are initialized */
    {
      for (i=0;i<MAXTIMENUM;i++)
	time_init(i);
      init = 1;
    }

  if( timelabel<0||timelabel>=MAXTIMENUM)
    {
      fprintf(stderr,"no such time label %d\n",timelabel);
      return(0);
    }
  if (time_flag(timelabel) != 1)
    {
      fprintf(stderr,
	      "no matching elapsetime_e call for previous elapsetime_s call\n%s.\n",Time_trans[timelabel]);
      return (0);
    }
  else
    {
      time_flag(timelabel) = 0;

      elapsetime(file,timelabel,print);

      thistime = runtime-time_stamp(timelabel);
      time_interval(timelabel) += thistime;
      time_stamp(timelabel) = runtime;
      if (print==1)
	fprintf(file,"%s TIME =%.4f\n",Time_trans[timelabel],thistime);
      return ((float) thistime);
    }
} /* elpasetime_e */

float elapsetime_m(FILE *file, int timelabel, int print)
{ /* elapsetime_e */
  float thistime;
  int i;

  if (init!=1) /* check whether the time structures are initialized */
    {
      for (i=0;i<MAXTIMENUM;i++)
	time_init(i);
      init = 1;
    }

  if( timelabel<0||timelabel>=MAXTIMENUM)
    {
      fprintf(stderr,"no such time label %d\n",timelabel);
      return(0);
    }
  if (time_flag(timelabel) != 1)
    {
      fprintf(stderr,
	      "no matching elapsetime_e call for previous elapsetime_s call\n%s.\n",Time_trans[timelabel]);
      return (0);
    }
  else
    {
      // time_flag(timelabel) = 0;

      elapsetime(file,timelabel,print);

      thistime = runtime-time_stamp(timelabel);
      time_interval(timelabel) += thistime;
      time_stamp(timelabel) = runtime;
      if (print==1)
	fprintf(file,"%s TIME =%.4f\n",Time_trans[timelabel],thistime);

      return ((float) time_interval(timelabel));
      //return ((float) thistime);
    }
} /* elpasetime_m */

void reset_time(int timelabel)
{
  time_interval(timelabel) = 0.0;
}

float get_time(int timelabel)
{

  return ((float) time_interval(timelabel));
}













