#include <stdio.h>
#include <stdlib.h>
#include <string.h>
/*************************************************************************
  levelorder.c -- Routines for assigning level numbers for gates and 
                  rearranging gates if they are not in level order 
*************************************************************************/
#include "Include/all.h"

/***************************************************************************
  copycircuit(node1, node2) copies contents of node1 to node2
***************************************************************************/
void copycircuit(node1, node2)
     NODE *node1;
     NODE *node2;
{ /* copycircuit */

  bcopy(node1, node2, sizeof(NODE));

} /* copycircuit */

/****************************************************************************
     s_enqueuefanouts() enqueues fanouts connected to a gate
     Inputs: gatenmbr --> gate number whose fanouts are to be queued.
     Function: Queue events to propagate the signal through all fanout gates
               connected to this gate.
*****************************************************************************/
void s_enqueuefanouts(gate)
     NODE *gate;
{ /* s_enqueuefanouts */
  int cfindex          /* Gate input index -- number I/P's - 1 */;

  for(cfindex = 0;cfindex < nmbroutputs (gate); ++cfindex)
    {
      s_enqueue(fanoutgate(gate, cfindex));
    }
} /* s_enqueuefanouts */

/************************************************************************
  Routine to assign new gate numbers 
  Loops through all the PI's, gates and PO's and assigns gate #'s 
  Assumes that the PI's, gates and PO's are already in levelorder.
************************************************************************/
void assign_gate_numbers(ckt)
     CIRCUIT *ckt;
{ /* assign_gate_numbers */
  int i;
  int gate_counter=1;
  int cfindex;
  NODE * ptr;

  for (i=0; i < nmbr_pi(ckt); i++)
    {
      ptr = PIs (ckt, i);
      gate_number(ptr) = gate_counter++;
    }
  for (i=0; i < nmbr_gates(ckt); i++)
    {
      gate_number(gatesx(ckt, i)) = gate_counter++;
    }
  for (i=0; i < nmbr_po(ckt); i++)
    {
      gate_number(POs(ckt, i)) = gate_counter++;
    }
  /* Get the comp_number fields in all fanin/fanout lists set right */
  for (i=0; i < nmbr_pi(ckt); i++)
    {
      ptr = PIs (ckt, i);
      for(cfindex = 0;cfindex < nmbroutputs (ptr); ++cfindex)
	{
	  fanoutgatenmbr (ptr, cfindex) =
	    gate_number (fanoutgate (ptr, cfindex));
	}
    }
  for (i=0; i < nmbr_gates(ckt); i++)
    {
      ptr = gatesx(ckt, i);
      for(cfindex = 0;cfindex < nmbroutputs (ptr); ++cfindex)
	{
	  fanoutgatenmbr (ptr, cfindex) =
	    gate_number (fanoutgate (ptr, cfindex));
	}
    }
  for (i=0; i < nmbr_po(ckt); i++)
    {
      ptr = POs(ckt, i);
      for(cfindex = 0;cfindex < nmbroutputs (ptr); ++cfindex)
	{
	  fanoutgatenmbr (ptr, cfindex) =
	    gate_number (fanoutgate (ptr, cfindex));
	}
    }
} /* assign_gate_numbers */


/************************************************************************
  assignlevelnumber() assigns the correct level numbers to the gates.
*************************************************************************/
void assignlevelnumber(CIRCUIT *circuit)
     //CIRCUIT *circuit;
{ /* assignlevelnumber */
  int i;
  NODE *igate;
  int l_level;
  
  /* Initialization */
  s_queueinit(QUEUE_SIZE);
  for(i=0;i<nmbr_pi(circuit);++i)
    {
      levelnumber(PIs(circuit, i)) = 1;
      /* queue fanouts of i */
      s_enqueuefanouts(PIs(circuit,i));
    }
  circuit->highest_level = 0;

  for(i=0;i<nmbr_gates(circuit);++i)
    {
      levelnumber(gatesx(circuit, i)) = -1;
    }

  for(i=0;i<nmbr_po(circuit);++i)
    {
      levelnumber(POs(circuit, i)) = -1;
    }

  while(s_queue_not_empty())
    {
      igate = s_dequeue();
      for(i=0;i<nmbrinputs(igate);++i)
	{
	  if((l_level = levelnumber(sourcegate(igate, i)))
	     == -1)
	    {
	      levelnumber(igate) = -1;
	      break;
	    }
	  levelnumber(igate) = 
	    maxm(l_level, levelnumber(igate));
	}

      if(levelnumber(igate) != -1)
	{
	  l_level = (levelnumber(igate) += 1);
	  if(gatefunction(igate) != p_po)
	    circuit->highest_level = maxm(l_level, circuit->highest_level);
	  
	  s_enqueuefanouts(igate);
	}
    }
  circuit->highest_level += 1;  /* to accomodate extra level for po's */
  
  for(i=0;i<nmbr_po(circuit);++i)
    levelnumber(POs(circuit, i)) = circuit->highest_level;
#ifdef RUDEBUG
  if(VERBOSE == TRUE)
    fprintf(resultsout, "highest_level = %d.\n", circuit->highest_level);
#endif
  for(i=0;i<nmbr_pi(circuit);++i)
    {
      if(levelnumber(PIs(circuit, i)) == -1)
	{
	  if(VERBOSE == TRUE)
	    fprintf(resultsout,
		    "assignlevelnumber: level # not assigned--PI indexed %d.\n",
		    i);
	  fatal_error("assignlevelnumber", COMPUTATION_ERROR, NULL);
	}
    }
  for(i=0;i<nmbr_gates(circuit);++i)
    {
      if(levelnumber(gatesx(circuit, i)) == -1)
	{
	  if(VERBOSE == TRUE)
	    fprintf(resultsout,
		  "assignlevelnumber: level # not assigned--gate indexed %d.\n",
		    i);
	  fatal_error("assignlevelnumber", COMPUTATION_ERROR, NULL);
	}
    }
} /* assignlevelnumber */


/*****************************************************************************
  Sort the gate list by level order
*****************************************************************************/
void sort_by_level(node_list, l, r)
     NODE *node_list[];
     int l, r;
{ /* sort_by_level */
  int i, j, k, m;
  NODE **tmp_node_list;

  tmp_node_list = (NODE **)calloc(r+1, sizeof(NODE *));

  if (r > l)
    {
      m = (r+l)/2;
      sort_by_level(node_list, l, m);
      sort_by_level(node_list, m+1, r);

      for (i = m+1; i > l; i--)
	tmp_node_list[i-1] = node_list[i-1];
      for (j = m; j < r; j++)
	tmp_node_list[r+m-j] = node_list[j+1];
      for (k = l; k <= r; k++)
	node_list[k] =
	  (levelnumber(tmp_node_list[i]) < levelnumber(tmp_node_list[j])) ?
	    tmp_node_list[i++] : tmp_node_list[j--];
    }

  free(tmp_node_list);
} /* sort_by_level */

/* update indicies of gates */
void update_gate_index(ckt)
     CIRCUIT *ckt;
{ /* update_gate_index */
  int i;
  
  for (i=0; i<nmbr_pi(ckt); i++)
    {
      gate_index(PIs(ckt, i)) = i;
    }

  for (i=0; i < nmbr_gates(ckt); i++)
    {
      gate_index(gatesx(ckt, i)) = i;
    }

  for (i=0; i < nmbr_po(ckt); i++)
    {
      gate_index(POs(ckt, i)) = i;
    }
} /* update_gate_index */

void arrange_in_levelorder(ckt)
     CIRCUIT *ckt;
{ /* arrange_in_levelorder */
  assignlevelnumber(ckt);
  sort_by_level(gate_list(ckt), 0, nmbr_gates(ckt)-1);
  update_gate_index(ckt);
} /* arrange_in_levelorder */
