#include "Include/all.h"
#ifdef STATSIM
#include "Include/haz.h"
#endif
//#include "Include/extern.h"

/* Directory of configuration options:
  You must define MAXSTRSIZE to be the correct maximum string length
  You must define NAME_WIDTH to be the correct name string length
  You must define SMALL_WIDTH to be the correct small string length
  Define STATSIM          to initialize currentoutputval2 and delay fields --
                          otherwise, they are not set
  Define ORIGGATENO       to remember original Rutgers gate number in netlist --
                          otherwise, it is forgotten
  Define FALSEDAT         to save delays in gate records -- 
                          otherwise, they are forgotten
  Define QUAD01           to build the weighted signed graph, the flipline structure,
                          and the adjacency matrix, read in the critical path file,
                          and initialize fanout edge weights to 1 --
                          otherwise, this is not done
  Define FALSEPATHS       to build the false timing paths data structure --
                          otherwise, it is not there
  Define ELABFALSE        to add the rise_true, fall_true, Cubes_Calc, Cubes, and ef
                          fields to the netlist data structure --
                          otherwise, they are not there
  Define COUNTFANOUTS     to count the number of fanouts in the circuit --
                          otherwise, it is not done
  Define CONVERT          to create suitable L or EDIF gate names from gate numbers --
                          otherwise, the name field is not set */
FLAG nonpi_seen;
FLAG po_seen;
FLAG num_dff;
FLAG seq_circuit;

char *errmsg1 = "Non-pi gate type must not precede pi on line %d\n";
char *errmsg2 = "Non-po gate type must not follow po on line %d\n";

/***************************************************************************
  search if tgate is a fanin of sgate
****************************************************************************/
int searchfanin(NODE *sgate, NODE *tgate)
{ /* searchfanin */
  int i,j;

  j = nmbrinputs (sgate);

  for(i=0;i<j;++i)
    {
      if(sourcegate(sgate,i) == tgate)
        return(i);
    }
  return(-1);
} /* searchfanin */

/****************************************************************************
     wrongformat() reports if the format is wrong in the given file.
     Inputs: line --> the line number.
             name --> FILE pointer to the file.
     Function: Print an error message to report the wrong format.
     Return Value: None.
 ****************************************************************************/

void wrongformat(int line, char *name)
{ /* wrongformat */
  fprintf(resultsout, "Illegal gatetype on line %d of schematic file %s.\n", line, name);
  fatal_error("wrongformat", INPUT_ERROR, NULL);
} /* wrongformat */

/****************************************************************************
     add_to_src() adds tgate as a fanout of sgate.  Information pertaining to
                  to the fanout gates is stored in a linked list "outputs",
		  and "outputindex", an integer, points to the next free
		  slot in the list where the information can be added.
     Inputs: sgate --> node number of source gate or the driving gate.
             tgate --> node number of the driven gate.
	     input_nmbr --> input number of tgate which is driven by sgate.
     Function: Add to the output list of sgate the node number of the tgate,
               and input number of the tgate which is connected to sgate.
	       Initialise the delay type of the line to unknown, and delay
	       size to zero.
     Side effects: Increment outputindex of sgate.
                   Sets the 'connected_to_po' if the tgate is a PO.
 ****************************************************************************/
void add_to_src(NODE *sgate, NODE *tgate, int inp_nmbr)
{ /* add_to_src */
  int j;
  
  j = outputindex(sgate);
  if(++outputindex(sgate) > nmbroutputs(sgate))
    {
      fprintf(resultsout, "add_to_src: Fanout list overflow -- gate %d.\n", gate_number(sgate));
      fatal_error("add_to_src", LIST_OVERFLOW, NULL);
    }
  
  fanoutgate(sgate,j) = tgate;
  fanoutgateinpnmbr(sgate,j) = inp_nmbr;

#ifdef POCONNET
  if(gatefunction(tgate) == p_po)
    node_connected_to_po(sgate) = TRUE;
#endif
} /* add_to_src */

/****************************************************************************/
/****************************************************************************/
char * r_trans (P_GATETYPE some_gate)
{ /* r_trans */
  switch (some_gate)
    {
    case p_and:      return ("and");
      break;
    case p_or:       return ("or");
      break;
    case p_nand:     return ("nand");
      break;
    case p_nor:      return ("nor");
      break;
    case p_xor:      return ("xor");
      break;
    case p_cxor:     return ("xor");
      break;
    case p_equiv:    return ("xnor");
      break;
    case p_not:      return ("inv");
      break;
    case p_d_ff:     return ("dff");
      break;
    case p_buffer:   return ("buff");
      break;
    case p_pi:       return ("pi");
      break;
    case p_po:       return ("po");
      break;
    case p_unknown:
    default:         return ("unknown");
      break;
    }
} /* r_trans */

/****************************************************************************
     r_parse () is the standard rut format parser, with slight modifications
               to handle delay information of gates.  The netlist format 
	       is as follows:
       <node_number> <gate_type> <input_list>

      Inputs: r_circuit -- pointer to a pointer to the circuit graph
                           (allocated by this function)
              schfile --> character pointer to the schematic file.
	      wantlevelsort --> set to TRUE if you want the netlist sorted
      Function: Parse the file in the above format. Report errors.
                Construct circuit graph.
      Side effects: Open the schfile and closes it after parsing it.
      Return value: An integer 1 if the circuit parsed and is sequential,
                               0 if the circuit parsed and is combinational,
                               < 0 if parsing failed
 ****************************************************************************/
int r_parse (CIRCUIT ** r_circuit,
	     char *schfile,        /* pointer to the file name */
	     int wantlevelsort)    /* Set to TRUE if you want level sort */
{ /* r_parse */
  int new_po_cnt;                  /* Count of # of PO's added */
  int maxgates = 0;
  FILE *sfp;
  char ginstr[NMBROFGATEINPUTS + 2][SMALL_WIDTH];
  char line[MAXSTRSIZE];
  int igate;
  char ifunc[3];
  int scnt;
  int origscnt;
  int i,j,k,l;
#ifdef QUAD01
  int jj;                     /* Loop index counter for initializing fanout
				 edge weights */
#endif
  int srcgate;
  FLAG GATESNOTINLEVELORDER = FALSE;
  int gatecount;
  NODE **tmp_gates;
  CIRCUIT *newcircuit;
#ifdef FALSEPATHS
  NODE *tmp_node;
#endif
  int gatedelay;       /* Integer gate delay from circuit schematic */
#ifdef CONVERT
  char temp [MAXSTRSIZE];
#endif
  NODE * gaterecord;
  int dummy_output_cnt;
  char po_port_name [MAXSTRSIZE];
  NODE * recip_comp_ptr;
  NODE * new_po;
  int jcount;
  NODE *somegate;
  int gate_cntr = 0;
  int dff_cntr = 0;
  int num_in;

  sfp = fopen (schfile, "r");
  if (sfp == NULL)
    {
      fatal_error ("r_parse", FILE_OPEN_FAILURE, schfile);
    }
  
  gatecount = 0;            /* Initialize the gate count */ 

  /* Calculating gate count for allocating space */ 
  while (fgets (line, 100, sfp) != NULL)
    {
      scnt = sscanf(line, "%d%2s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
		    &igate, ifunc, 
		    ginstr[0], ginstr[1], ginstr[2], ginstr[3], ginstr[4],
		    ginstr[5], ginstr[6], ginstr[7], ginstr[8], ginstr[9],
		    ginstr[10], ginstr[11], ginstr[12], ginstr[13],
		    ginstr[14], ginstr[15], ginstr[16]);
      if (scnt > 0)
	gatecount ++;
    }

  fclose (sfp); /* Close the file since we will be rereading it from the
		   beginning */
  gatecnt = gatecount;/* This is the total # of gates including PIs, POs and FFs -- OK */
  if(gatecount < 3)
    fatal_error ("r_parse", EMPTY_INPUT_FILE, NULL);

  newcircuit = (CIRCUIT *)calloc(1,sizeof(CIRCUIT));
  //gate_list(newcircuit)
  tmp_gates = (NODE **) calloc(gatecount, sizeof(NODE *));
  wantlevelsort = TRUE;
  /* Initialization  */
  
  for(i=0; i<gatecount; ++i)
    {
      tmp_gates[i] = (NODE *)calloc(1, sizeof(NODE));
      nmbroutputs (node_of_list(tmp_gates,i)) = 0;
      outputindex(node_of_list(tmp_gates,i)) = 0;
      fanouts (node_of_list(tmp_gates,i)) = NULL;
      
#ifdef QUAD01
      adjoutput (node_of_list(tmp_gates, i)) = NULL;
#endif
#ifdef STATSIM
#ifdef QUAD01
      currentoutputval2 (node_of_list (tmp_gates, i)) = x_d_x;
#else
      currentoutputval (node_of_list (tmp_gates, i)) = x_d_x;
#endif
      rising_gatedel (node_of_list (tmp_gates, i)) = 0;
      falling_gatedel (node_of_list (tmp_gates, i)) = 0;
#endif
      nmbrinputs(node_of_list(tmp_gates, i)) = 0;
#ifdef QUAD01
      number_flippers_on_inputs(node_of_list(tmp_gates,i)) = 0;
#endif
      for(j=0;j</*nmbrinputs(node_of_list(tmp_gates,i))*/tmp_gates[i]->nmbr_inputs;++j)
	{
	  sourcegate(node_of_list(tmp_gates, i),j) = NULL;
	}
    }

  sfp = fopen (schfile, "r");
  if (sfp == NULL)
    {
      fatal_error ("r_parse", FILE_OPEN_FAILURE, schfile);
    }
  
  nonpi_seen = FALSE;
  po_seen = FALSE;
  num_dff = 0;
  seq_circuit = 0;
  nmbr_pi(newcircuit) = 0;
  nmbr_po(newcircuit) = 0;
  
  while(fgets(line, MAXSTRSIZE, sfp) != NULL)
    {
      /* while more input */
      maxgates++;
      if(maxgates > gatecount)
	{
	  fprintf(resultsout,
		  "r_parse: circuit graph overrun -- %d.\n", gatecount);
	  fatal_error("r_parse", CIRCUIT_GRAPH_OVERRUN, NULL);
	}
      scnt = sscanf(line, "%d%2s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
		    &igate, ifunc, 
		    ginstr[0], ginstr[1], ginstr[2], ginstr[3], ginstr[4],
		    ginstr[5], ginstr[6], ginstr[7], ginstr[8], ginstr[9],
		    ginstr[10], ginstr[11], ginstr[12], ginstr[13],
		    ginstr[14], ginstr[15], ginstr[16]);

      gate_number(tmp_gates[igate-1]) = igate;
#ifdef ORIGGATENO
      tmp_gates[igate-1]->orig_gate_nmbr = igate;
#endif
      switch(ifunc[0])
	{ /* switch(ifunc[0]) */
	case 'p':
	  switch(ifunc[1])
	    {
	    case 'i':
	      gatefunction (node_of_list(tmp_gates, igate-1)) = p_pi;
	      if(nonpi_seen == TRUE)
		fprintf(resultsout, errmsg1, maxgates);
	      if(po_seen == TRUE)
		fprintf(resultsout, errmsg2, maxgates);
	      nmbr_pi(newcircuit)++;
	      num_pi++;/* The Tool uses this variable everywhere */
	      break;
	    case 'o':
	      gatefunction (node_of_list(tmp_gates, igate-1)) = p_po;
	      nonpi_seen = TRUE;
	      if(po_seen != TRUE)
		{
		  po_seen = TRUE;
		}
	      nmbr_po(newcircuit)++;
	      num_po++; /* The Tool uses this variable everywhere */
	      break;
	    default:
	      wrongformat(maxgates, schfile);
	      break;
	    }
	  break;
	case 'a':
	  switch(ifunc[1])
	    {
	    case 'n':
	      gatefunction (node_of_list(tmp_gates, igate-1)) = p_and;
	      nonpi_seen = TRUE;
	      if(po_seen == TRUE)
		fprintf(resultsout, errmsg2, maxgates);
	      break;
	    default:
	      wrongformat(maxgates, schfile);
	      break;
	    }
	  break;
	case 'c':
	  switch(ifunc[1])
	    {
	    case 'x':
	      gatefunction (node_of_list(tmp_gates, igate-1)) = p_cxor;
	      nonpi_seen = TRUE;
	      if(po_seen == TRUE)
		fprintf(resultsout, errmsg2, maxgates);
	      break;
	    default:
	      wrongformat(maxgates, schfile);
	      break;
	    }
	  break;
     	case 'o':
	  switch(ifunc[1])
	    {
	    case 'r':
	      gatefunction (node_of_list(tmp_gates, igate-1)) = p_or;
	      nonpi_seen = TRUE;
	      if(po_seen == TRUE)
		fprintf(resultsout, errmsg2, maxgates);
	      break;
	    default:
	      wrongformat(maxgates, schfile);
	      break;
	    }
	  break;
	case 'n':
	  switch(ifunc[1])
	    {
	    case 'a':
	      gatefunction (node_of_list(tmp_gates, igate-1)) = p_nand;
	      nonpi_seen = TRUE;
	      if(po_seen == TRUE)
		fprintf(resultsout, errmsg2, maxgates);
	      break;
	    case 'o':
	      gatefunction (node_of_list(tmp_gates, igate-1)) = p_nor;
	      nonpi_seen = TRUE;
	      if(po_seen == TRUE)
		fprintf(resultsout, errmsg2,maxgates);
	      break;
	    default:
	      wrongformat(maxgates, schfile);
	      break;
	    }
	  break;
        case 'i':
          switch(ifunc[1])
            {
	    case 'n':
	      gatefunction (node_of_list(tmp_gates, igate-1)) = p_not;
	      nonpi_seen = TRUE;
	      if(po_seen == TRUE)
	        fprintf(resultsout, errmsg2, maxgates);
	      break;
	    default:
	      wrongformat(maxgates, schfile);
	      break;
	    }
	  break;
        case 'x':
          switch(ifunc[1])
            {
	    case 'o':
	      gatefunction (node_of_list(tmp_gates, igate-1)) = p_xor;
#ifdef QUAD01
	      nmbr_xorgates(newcircuit)++;
#endif
	      nonpi_seen = TRUE;
	      if(po_seen == TRUE)
	        fprintf(resultsout, errmsg2, maxgates);
	        break;
	    default:
	      wrongformat(maxgates, schfile);
	      break;
	    }
	  break;
        case 'e':
	  switch(ifunc[1])
	    {
	    case 'q':
	      gatefunction (node_of_list(tmp_gates, igate-1)) = p_equiv;
#ifdef QUAD01	      
	      nmbr_xorgates(newcircuit)++;
#endif
	      nonpi_seen = TRUE;
	      if(po_seen == TRUE)
	        fprintf(resultsout, errmsg2, maxgates);
	      break;
	    default:
	      wrongformat(maxgates, schfile);
	      break;
	    }
	  break;
	case 'd':
	  switch (ifunc[1])
	    {
	    case 'f':
	      gatefunction (node_of_list(tmp_gates, igate-1)) = p_d_ff;
	      num_dff ++;
	      num_pi++;// num_pi is actually # of PIs + # of DFFs
	      num_po++;// num_po is actually # of POs + # of DFFs
	      gatecnt++; // gatecnt is actually # of Total Gates (LG,FFs + POs + PIs) + # of DFFs
	      nonpi_seen = TRUE;
	      seq_circuit = TRUE;
	      if (po_seen == TRUE)
		fprintf(resultsout, 
			"Non-po gate type must not follow po on line %d\n", 
			maxgates);
	      break;
	    default:
	      wrongformat(maxgates, schfile);
	      break;
	    }
	  break;
        default:
          wrongformat(maxgates, schfile);
          break;	
	} /* switch(ifunc[0]) */
      fflush (resultsout);

      /* Insert the rising/falling gate delays into the data structure */
      origscnt = scnt;
      if(scnt > 2)
	{
          if (gatefunction (node_of_list (tmp_gates, igate - 1)) == p_po)
	    {
	      j = scnt-2;
	    }
          else
            {
	    restart:
	      switch(ginstr[scnt-3][0])
	        {
	        case 'd':
		  j = scnt - 3;
	          jcount = sscanf(&ginstr[j][1], "%d", &gatedelay);
		  if (jcount <= 0)
		    {
		      j = scnt - 2;
		      jcount = sscanf(&ginstr[j][0], "%d", &gatedelay);
		    }
		  -- scnt;
#ifdef FALSEDAT
	          rising_gatedel (node_of_list (tmp_gates, igate - 1)) = gatedelay;
	          falling_gatedel (node_of_list (tmp_gates, igate - 1)) = gatedelay;
#endif
	          break;
		case 'f':
		  j = scnt - 3;
	          jcount = sscanf(&ginstr[j][1], "%d", &gatedelay);
		  if (jcount <= 0)
		    {
		      j = scnt - 2;
		      jcount = sscanf(&ginstr[j][0], "%d", &gatedelay);
		    }
		  -- scnt;
#ifdef FALSEDAT
	          falling_gatedel (node_of_list (tmp_gates, igate - 1)) = gatedelay;
#endif
		  j = scnt - 3;
		  if(ginstr[j][0] != 'r')
		    {
		      j = scnt - 4;
		      if(ginstr[j][0] != 'r')
			{
			  fprintf(resultsout,
				  "r_parse: Improper delay assignment--line %d.\n", maxgates);
			  fflush (resultsout);
			  exit(-5);
			}
		      jcount = sscanf(&ginstr[j + 1][0], "%d", &gatedelay);
		      -- scnt;
		    }
		  else
		    jcount = sscanf(&ginstr[j][1], "%d", &gatedelay);
		  -- scnt;
#ifdef FALSEDAT
	          rising_gatedel (node_of_list (tmp_gates, igate - 1)) = gatedelay;
#endif
	          break;
		case 'r':
		  j = scnt - 3;
	          jcount = sscanf(&ginstr[j][1], "%d", &gatedelay);
		  if (jcount <= 0)
		    {
		      j = scnt - 2;
		      jcount = sscanf(&ginstr[j][0], "%d", &gatedelay);
		    }
		  -- scnt;
#ifdef FALSEDAT
	          rising_gatedel (node_of_list (tmp_gates, igate - 1)) = gatedelay;
#endif
		  j = scnt - 3;
		  if(ginstr[j][0] != 'f')
		    {
		      j = scnt - 4;
		      if(ginstr[j][0] != 'f')
			{
			  fprintf(resultsout,
				  "r_parse: Improper delay assignment--line %d.\n", maxgates);
			  fflush (resultsout);
			  exit(-5);
			}
		      jcount = sscanf(&ginstr[j + 1][0], "%d", &gatedelay);
		      -- scnt;
		    }
		  else
		    jcount = sscanf(&ginstr[j][1], "%d", &gatedelay);
		  -- scnt;
#ifdef FALSEDAT
	          falling_gatedel (node_of_list (tmp_gates, igate - 1)) = gatedelay;
#endif
	          break;
	        default:
		  -- scnt;
		  if (scnt >= 2) goto restart;
		  scnt = origscnt;
	          break;
		}
	    }
	}

      switch(gatefunction(node_of_list(tmp_gates, igate-1)))
	{ /* switch begin */
	case p_pi:
	  if (scnt == 2)
	    nmbrinputs(node_of_list(tmp_gates, igate-1)) = 0;
	  else
	    wrongformat(maxgates, schfile);
	  break;
	case p_po:
	case p_not:
	  if(scnt == 3)
	    {
	      nmbrinputs(node_of_list(tmp_gates, igate-1)) = 1;
	      fanins(node_of_list(tmp_gates, igate-1)) =
		(FANIN *)calloc(nmbrinputs(node_of_list(tmp_gates, igate-1)), sizeof(NODE *));
	      /* check of gates not in Level Order */
	      if((srcgate = atoi(ginstr[0])) > igate)
		{
		  GATESNOTINLEVELORDER = TRUE;
		}
	      /* check if input of a gate shorted to its output */
	      else if(srcgate == igate)
		{
		  fprintf(resultsout, "output of gate %d shorted to input.\n",
			  igate);
		  fatal_error("r_parse", INPUT_ERROR, NULL);
		}
	      else if(srcgate <= 0)
		{
		  fprintf(resultsout, "source gate of %d is %d Illegal.\n", igate,
			  srcgate);
		  fatal_error("r_parse", INPUT_ERROR, NULL);
		}
	      sourcegate(node_of_list(tmp_gates, igate-1), 0) = tmp_gates[srcgate-1];
	      nmbroutputs (node_of_list(tmp_gates,srcgate-1)) += 1;  
	    }
	  else
	    wrongformat(maxgates, schfile);
	  break;
	default:
	  nmbrinputs(node_of_list(tmp_gates, igate-1)) = scnt-2;
	  fanins(node_of_list(tmp_gates, igate-1)) = (FANIN *)calloc(nmbrinputs(node_of_list(tmp_gates, igate-1)),
					      sizeof(NODE *));
	  for(l=0,i=0; i<nmbrinputs(node_of_list(tmp_gates,igate-1)); ++i)
	    {
	      /* check of gates not in Level Order */
	      if((srcgate = atoi(ginstr[i])) > igate)
		GATESNOTINLEVELORDER = TRUE;
	      /* check if input of a gate shorted to its output */
	      else if(srcgate == igate)
		{
		  fprintf(resultsout, "output of gate %d shorted to input.\n",
			  igate);
		  fatal_error("r_parse", INPUT_ERROR, NULL);
		}

	      if(srcgate <= 0)
		{
		  fprintf(resultsout,"source gate of %d is %d. \n",igate,srcgate);
		  fatal_error("r_parse", INPUT_ERROR, NULL);
		}

	      /* check if a po is shorted to a pi directly */
	      if(gatefunction(node_of_list(tmp_gates,igate-1)) == p_po)
		{
		  if(gatefunction(node_of_list(tmp_gates,srcgate-1)) == p_pi)
		    {
		      fprintf(resultsout, "WARNING: PI %d is shorted to PO %d.\n",
			      igate, srcgate);
		      fflush(resultsout);
		    }
		}
	      /* check if inputs are duplicated */
	      if(l>0)
		for(k=0;k<l;++k)
		  {
		    if(sourcegate(node_of_list(tmp_gates,igate-1),k) ==
		       node_of_list(tmp_gates,srcgate-1))
		      {
			fprintf(resultsout, "WARNING: duplicated inputs--gate%d.\n",
				igate);
			fflush (resultsout);
			srcgate = 0;
			--nmbrinputs(node_of_list(tmp_gates,igate-1));
			break;
		      }
		    
		  }
	      if(srcgate != 0) /* avoid recording duplicated inputs */
		{
		  sourcegate(node_of_list(tmp_gates, igate-1), l++) = tmp_gates[srcgate-1];
		  nmbroutputs (node_of_list(tmp_gates,srcgate-1)) += 1;  
		}

	      if ((gatefunction (node_of_list(tmp_gates, igate-1)) == p_d_ff)
		  && (scnt - 2 > 1))
		{
		  fprintf(resultsout, "Error: More than one input to D-Flip-Flop %d.\n",
			 igate);
		  fprintf(resultsout,
			  "Check with the rutools manual. Aborted.\n");
		  fflush (resultsout);
		  exit(-1);
		}
	    }
	} /* switch end */
    } /* endwhile(fgets(line, MAXSTRSIZE, sfp) != NULL) */

  /* Close the schematic file */
  fclose(sfp);
  
  /* Now allocate individual lists for PI's, PO's, PPOs and gates */
  pi_list(newcircuit) = (NODE **) calloc(nmbr_pi(newcircuit), sizeof(NODE *));
  nmbr_gates(newcircuit) = gatecount-nmbr_pi(newcircuit)-nmbr_po(newcircuit);
  gate_list(newcircuit) = (NODE **) calloc(nmbr_gates(newcircuit), sizeof(NODE *));
  po_list(newcircuit) = (NODE **) calloc(nmbr_po(newcircuit), sizeof(NODE *));
  get_po_alloc(newcircuit) = nmbr_po (newcircuit);
  get_pi_alloc(newcircuit) = nmbr_pi (newcircuit);
  get_gate_alloc(newcircuit) = nmbr_gates (newcircuit);
  //newcircuit->Ppo_list = (NODE **)calloc(num_dff, sizeof(NODE *));
  ppo_list(newcircuit) = (NODE **)calloc(num_dff, sizeof(NODE *));

  for (i=0; i<nmbr_pi(newcircuit); i++)
    {
      /* Allocating Memory for PI List of New Circuit */
      PIs(newcircuit, i) = (NODE *)calloc(1, sizeof(NODE));
      PIs(newcircuit, i) = tmp_gates[i];
      gate_index(PIs(newcircuit, i)) = i;
      gate_number(PIs(newcircuit,i)) = tmp_gates[i]->gate_nmbr;/* OK */
      /* More members of the structure NODE can be added here -- OK */
#ifdef CONVERT
      sprintf (temp, "pi%d", gate_number (PIs (newcircuit, i)));
      compname (PIs (newcircuit, i)) = (char *) strcreate (temp);
#endif
    }
  /* Initializing PPO List for Circuit */
  for (i=0; i < num_dff; i++)
    {
      newcircuit->Ppo_list[i] = (NODE *)calloc(1, sizeof(NODE));
    }

  dff_cntr = 0;
  gate_cntr = num_dff;
  for (i=0; i < nmbr_gates(newcircuit); i++)
    {
      gatesx(newcircuit, i) = (NODE *)calloc(1, sizeof(NODE));
/*        if(tmp_gates[i+nmbr_pi(newcircuit)]->funct == p_d_ff) */
/*  	{ */
/*  	  if(dff_cntr >= num_dff) */
/*  	    { */
/*  	      printf("ERROR -- Wrong index # for FF while copying in Gate List!\n"); */
/*  	      exit(-1); */
/*  	    } */
/*  	  gatesx(newcircuit, dff_cntr++) = tmp_gates[i+nmbr_pi(newcircuit)]; */
/*  	} */
/*        else */
/*  	{ */
/*  	  if(gate_cntr < num_dff || gate_cntr> nmbr_gates(newcircuit)) */
/*  	    { */
/*  	      printf("ERROR -- Wrong index # for Gate while copying in Gate List!\n"); */
/*  	      exit(-1); */
/*  	    } */
/*  	  gatesx(newcircuit, gate_cntr++) = tmp_gates[i+nmbr_pi(newcircuit)]; */
/*  	} */
      gatesx(newcircuit, i) = tmp_gates[i+nmbr_pi(newcircuit)];
      gate_index(gatesx(newcircuit, i)) = i;/* This looks useless! *//* OK */
      gate_number(gatesx(newcircuit,i)) = tmp_gates[i+nmbr_pi(newcircuit)]->gate_nmbr;/* OK */
      /* More members of the structure NODE can be added here -- OK */
      
#ifdef CONVERT
      sprintf (temp, "gate%d", gate_number (gatesx (newcircuit, i)));
      compname (gatesx (newcircuit, i)) = (char *) strcreate (temp);
#endif
    }

  for (i=0; i < nmbr_po(newcircuit); i++)
    {
      POs(newcircuit, i) = (NODE *)calloc(1, sizeof(NODE));
      POs(newcircuit, i) = tmp_gates[i+nmbr_pi(newcircuit)+nmbr_gates(newcircuit)];
      gate_index(POs(newcircuit, i)) = i;
      gate_number(POs(newcircuit,i)) = tmp_gates[i+nmbr_pi(newcircuit)+nmbr_gates(newcircuit)]->gate_nmbr;
      /* More members of the structure NODE can be added here -- OK */
#ifdef CONVERT
      sprintf (temp, "po%d", gate_number (POs (newcircuit, i)));
      compname (POs (newcircuit, i)) = (char *) strcreate (temp);
#endif
    }
  
#ifdef COUNTFANOUTS
  /* Count the number of fanouts on PIs */
  for(i=0;i<nmbr_pi(newcircuit);++i)
    {
      if(nmbroutputs(PIs(newcircuit,i)) >= 2)
	{
	  ++nmbr_fanouts(newcircuit);
	}
    }

  /* Count the number of fanouts on gates */
  for(i=0;i<nmbr_gates(newcircuit);++i)
    {
      if(nmbroutputs(gatesx(newcircuit,i)) >= 2)
	{
	  ++nmbr_fanouts(newcircuit);
	}
    }
#endif

  for(i=0;i<nmbr_pi(newcircuit);++i)
    {
      if ((fanouts (PIs(newcircuit, i)) = 
	  (FANOUT *)calloc(nmbroutputs (PIs(newcircuit, i)),
			  sizeof(FANOUT))) == NULL)
	{
	  fatal_error("r_parse", MEMORY_ALLOC_FAILURE, NULL);
	}
#ifdef QUAD01
      else
	{
	  /* Initialize all fanout edge weights to 1 */
	  for (jj = 0; jj < nmbroutputs (PIs(newcircuit, i)); ++ jj)
	    fanoutedgeweight (PIs (newcircuit, i), jj) = 1;
	}
#endif
    }

  for(i=0;i<nmbr_gates(newcircuit);++i)
    {
      if ((fanouts (gatesx(newcircuit, i)) = 
	  (FANOUT *)calloc(nmbroutputs (gatesx(newcircuit,i)), sizeof(FANOUT))) == NULL)
	{
	  fatal_error("r_parse", MEMORY_ALLOC_FAILURE, NULL);
	}
#ifdef QUAD01
      else
	{
	  /* Initialize all fanout edge weights to 1 */
	  for (jj = 0; jj < nmbroutputs (gatesx (newcircuit, i)); ++ jj)
	    fanoutedgeweight (gatesx (newcircuit, i), jj) = 1;
	}
#endif
    }


  /* Now update the output lists of each gate that drives another gate */
  for(i=0;i<nmbr_gates(newcircuit);++i)
    {
      for(j=0;j<nmbrinputs(gatesx(newcircuit, i)); j++)
	{
	  add_to_src(sourcegate(gatesx(newcircuit, i),j), gatesx(newcircuit, i), j);
	}
    }

  /* Now update the output lists of each gate driving the POs */
  for(i=0;i<nmbr_po(newcircuit);++i)
    {
      add_to_src(sourcegate(POs(newcircuit, i), 0), POs(newcircuit, i), 0);
    }

  /*Before we call 'arrange_in_levelorder', we duplicate the Gate pointers. We can use 
    duplicate list to find the gate pointers using gate number. Original order of gate 
    pointers is altered by 'arrange_in_levelorder' according to level number and not 
    gate number. */
  duplicate_gate_pntrs(newcircuit, nmbr_gates(newcircuit));

  /* queue.c does not need a queue greater that O(gatecount) as the
     queue_flag keeps track whether a gate is already queued or not */
  QUEUE_SIZE = nmbr_gates(newcircuit)+nmbr_po(newcircuit)+1;
  s_queueinit (QUEUE_SIZE);  /*  Allocate the Queue */

#ifdef QUAD01
  /* NOTE: WSGRAPH_SIZE should also include fanout nodes -- future */
  WSGRAPH_SIZE = nmbr_gates(newcircuit)+nmbr_pi(newcircuit)+1;
#endif
  printf("Goint into Level Sorting -- OK\n");
  if ((wantlevelsort == TRUE) && (GATESNOTINLEVELORDER == TRUE))
    {
      printf("Level Sorting the Gates -- OK\n");
      fprintf(resultsout,
	    "WARNING: Gates not in Level Order.\nArranging in level order...");
      fflush(resultsout);
      printf("Calling 'verify_ckt_topology(newcircuit)' first -- OK\n");
      verify_ckt_topology(newcircuit);
      printf("Returning from 'verify_ckt_topology(newcircuit)' -- OK\n");
      printf("Calling 'arrange_in_levelorder(newcircuit)' -- OK\n");
      arrange_in_levelorder(newcircuit);
      printf("Returning 'arrange_in_levelorder(newcircuit)' -- OK\n");
      fprintf(resultsout, "done\n");
      fflush(resultsout);
    }
  else
    {
      printf("Assigning Level Numbers to  the Gates -- OK\n");
      if(VERBOSE == TRUE) fprintf(resultsout, "Assigning level numbers...\n");
      printf("Calling 'assignlevelnumber(newcircuit)' -- OK\n");
      assignlevelnumber(newcircuit);
      printf("Returning from 'assignlevelnumber(newcircuit)' -- OK\n");
    }

#ifdef FALSEPATHS  
  /* Now construct the output list of each pi and gate */
  /* First allocate space for the output lists */
  for(i=0;i<nmbr_pi(newcircuit);++i)
    {
      /* Allocate data structures for false timing path calculation */
      /* Create redundant path count structure */
      tmp_node = PIs(newcircuit, i);
      tmp_node->rise_paths = (float *) calloc (nmbr_pi(newcircuit), sizeof (float));
      tmp_node->fall_paths = (float *) calloc (nmbr_pi(newcircuit), sizeof (float));
      tmp_node->rise_prop  = matrix_2d (nmbr_pi(newcircuit), nmbrinputs (tmp_node));
      tmp_node->fall_prop  = matrix_2d (nmbr_pi(newcircuit), nmbrinputs (tmp_node));
#ifdef ELABFALSE
      tmp_node->rise_true  = matrix_2d (nmbr_pi(newcircuit),
					nmbrinputs (tmp_node));
      tmp_node->fall_true  = matrix_2d (nmbr_pi(newcircuit),
					nmbrinputs (tmp_node));
#endif
      tmp_node->true_val   = matrix_2d (nmbr_pi(newcircuit), 4);
#ifdef ELABFALSE
      tmp_node->Cubes_Calc = (int *) callocate  (2, sizeof (int));
      tmp_node->Cubes = (float *) callocate (2, sizeof (float));
      tmp_node->ef = (EF_CUBE **) callocate (2, sizeof (EF_CUBE *));
#endif
    }
  for(i=0;i<nmbr_gates(newcircuit);++i)
    {
      /* Allocate data structures for false timing path calculation */
      /* Create redundant path count structure */
      tmp_node = gatesx(newcircuit, i);
      tmp_node->rise_paths = (float *) calloc (nmbr_pi(newcircuit), sizeof (float));
      tmp_node->fall_paths = (float *) calloc (nmbr_pi(newcircuit), sizeof (float));
      tmp_node->rise_prop  = matrix_2d (nmbr_pi(newcircuit), nmbrinputs (tmp_node));
      tmp_node->fall_prop  = matrix_2d (nmbr_pi(newcircuit), nmbrinputs (tmp_node));
#ifdef ELABFALSE
      tmp_node->rise_true  = matrix_2d (nmbr_pi(newcircuit),
					nmbrinputs (tmp_node));
      tmp_node->fall_true  = matrix_2d (nmbr_pi(newcircuit),
					nmbrinputs (tmp_node));
#endif
      tmp_node->true_val   = matrix_2d (nmbr_pi(newcircuit), 4);
#ifdef ELABFALSE
      tmp_node->Cubes_Calc = (int *) callocate  (2, sizeof (int));
      tmp_node->Cubes = (float *) callocate (2, sizeof (float));
      tmp_node->ef = (EF_CUBE **) callocate (2, sizeof (EF_CUBE *));
#endif
    }
  
  for (i=0; i<nmbr_po(newcircuit); i++)
    {
      tmp_node = POs(newcircuit, i);
      tot_paths(tmp_node) = 1.0;
      /* Allocate data structures for false timing path calculation */
      /* Create redundant path count structure */
      tmp_node->rise_paths = (float *) calloc (nmbr_pi(newcircuit), sizeof (float));
      tmp_node->fall_paths = (float *) calloc (nmbr_pi(newcircuit), sizeof (float));
      tmp_node->rise_prop  = matrix_2d (nmbr_pi(newcircuit), nmbrinputs (tmp_node));
      tmp_node->fall_prop  = matrix_2d (nmbr_pi(newcircuit), nmbrinputs (tmp_node));
#ifdef ELABFALSE
      tmp_node->rise_true  = matrix_2d (nmbr_pi(newcircuit),
					nmbrinputs (tmp_node));
      tmp_node->fall_true  = matrix_2d (nmbr_pi(newcircuit),
					nmbrinputs (tmp_node));
#endif
      tmp_node->true_val   = matrix_2d (nmbr_pi(newcircuit), 4);
#ifdef ELABFALSE
      tmp_node->Cubes_Calc = (int *) callocate  (2, sizeof (int));
      tmp_node->Cubes = (float *) callocate (2, sizeof (float));
      tmp_node->ef = (EF_CUBE **) callocate (2, sizeof (EF_CUBE *));
#endif
    }
#endif

  free(tmp_gates);

  verify_ckt_topology(newcircuit);

  new_po_cnt = 0;
  for (i = 0; i < nmbr_pi (newcircuit); ++ i)
    {
      gaterecord = PIs (newcircuit, i);
      if (nmbrinputs (gaterecord) > 0)
	{
#ifdef CONVERT
	  fprintf (resultsout,
		  "Gate %s is a pi gate with more than 0 inputs -- Exiting.\n",
		   compname (gaterecord));
#else
	  fprintf (resultsout,
		  "Gate %d is a pi gate with more than 0 inputs -- Exiting.\n",
		   gate_number (gaterecord));
#endif
	  fflush (resultsout);
	  exit (-1);
	}
      if (nmbroutputs (gaterecord) == 0)
	{
#ifdef CONVERT
	  fprintf (resultsout,
	   "Gate %s is a pi gate with no outputs -- making a new PO for it.\n",
		   compname (gaterecord));
#else
	  fprintf (resultsout,
	   "Gate %d is a pi gate with no outputs -- making a new PO for it.\n",
		   gate_number (gaterecord));
#endif
	  /* Make a new primary output and wire the PI logic to that */
	  dummy_output_cnt = nmbr_po (newcircuit);
	  sprintf (po_port_name, "DUMMYPO%d", dummy_output_cnt);
	  recip_comp_ptr = make_new_po (newcircuit, dummy_output_cnt,
					po_port_name,
					0);
	  fflush (resultsout);
	  add_wire (gaterecord, recip_comp_ptr);
	}
    }

  for (i = 0; i < nmbr_gates (newcircuit); ++ i)
    {
      gaterecord = gatesx (newcircuit, i);
      if (nmbroutputs (gaterecord) == 0) 
	{
#ifdef CONVERT
	  sprintf (po_port_name, "%s.dummypo%d", compname (gaterecord),
		   ++ new_po_cnt);
	  fprintf (resultsout,
   "Gate %s is a %s gate with no outputs -- adding dummy PO called %s.\n",
		   compname (gaterecord),
		   r_trans (gatefunction (gaterecord)), po_port_name);
#else
	  sprintf (po_port_name, "%d.dummypo%d", gate_number (gaterecord),
		   ++ new_po_cnt);
	  fprintf (resultsout,
   "Gate %d is a %s gate with no outputs -- adding dummy PO called %s.\n",
		   gate_number (gaterecord),
		   r_trans (gatefunction (gaterecord)), po_port_name);
#endif
	  fflush (resultsout);
	  new_po =  make_new_po (newcircuit, nmbr_po (newcircuit), po_port_name,
				 levelnumber (gaterecord) + 1);
	  add_wire (gaterecord, new_po);
	}
      switch (gatefunction (gaterecord))
	{
        case p_not:
	  if (nmbrinputs (gaterecord) != 1)
	    {
#ifdef CONVERT
	      fprintf (resultsout,
   "Error: Your inverter %s has %d inputs, but it should have 1 -- Exiting.\n",
		       compname (gaterecord), nmbrinputs (gaterecord));
#else
	      fprintf (resultsout,
   "Error: Your inverter %d has %d inputs, but it should have 1 -- Exiting.\n",
		       gate_number (gaterecord), nmbrinputs (gaterecord));
#endif
	      fflush (resultsout);
	      exit (-1);
	    }
	  break;
	default:
	  break;
	}
      if (nmbrinputs (gaterecord) == 0)
	{
#ifdef CONVERT
	  fprintf (resultsout,
		   "Gate %s is a %s gate with no inputs -- Exiting.\n",
		   compname (gaterecord),
		   r_trans (gatefunction (gaterecord)));
#else
	  fprintf (resultsout,
		   "Gate %d is a %s gate with no inputs -- Exiting.\n",
		   gate_number (gaterecord),
		   r_trans (gatefunction (gaterecord)));
#endif
	  fflush (resultsout);
	  exit (-1);
	}
      if (nmbrinputs (gaterecord) > 9)
	{
#ifdef CONVERT
	  fprintf (resultsout,
		   "Error: Gate %s has more than 9 inputs -- Exiting.\n",
		   compname (gaterecord));
#else
	  fprintf (resultsout,
		   "Error: Gate %d has more than 9 inputs -- Exiting.\n",
		   gate_number (gaterecord));
#endif
	  fflush (resultsout);
	  exit(-1);
	}
    }

  for (i = 0; i < nmbr_po (newcircuit); ++ i)
    {
      gaterecord = POs (newcircuit, i);
      if (nmbrinputs (gaterecord) != 1)
	{
#ifdef CONVERT
	  fprintf (resultsout,
   "Gate %s is a po gate with %d inputs (which should be 1) -- Exiting.\n",
		   compname (gaterecord), nmbrinputs (gaterecord));
#else
	  fprintf (resultsout,
   "Gate %d is a po gate with %d inputs (which should be 1) -- Exiting.\n",
		   gate_number (gaterecord), nmbrinputs (gaterecord));
#endif
	  fflush (resultsout);
	  exit (-1);
	}
      if (nmbroutputs (gaterecord) != 0) 
	{
#ifdef CONVERT
	  fprintf (resultsout,
		   "Gate %s is a po gate with outputs -- Exiting.\n",
		   compname (gaterecord));
#else
	  fprintf (resultsout,
		   "Gate %d is a po gate with outputs -- Exiting.\n",
		   gate_number (gaterecord));
#endif
	  fflush (resultsout);
	  exit (-1);
	}
    }

  j = 0;
  for(i=0; i<nmbr_gates(newcircuit); i++)
    {

      /* Copying FF Info to PPO List */
      somegate = gatesx_dup(newcircuit, i);
      if(gatefunction(somegate)==p_d_ff)
	{
	  /* Attach PPO to this DFF */
	  //somegate->ppo_list.gatepntr = PPOs(newcircuit, j);
	  /* Gate Number */
	  gate_number(PPOs(newcircuit, j)) = j + nmbr_gates(newcircuit)+nmbr_pi(newcircuit)+nmbr_po(newcircuit) + 1;
	  /* Gate Function */
	  gatefunction(PPOs(newcircuit, j)) = p_ppo;
	  /*FANIN *inputlist */
	  if(fanins(PPOs(newcircuit, j)) == NULL)
	    {
	      fanins(PPOs(newcircuit, j)) = (FANIN *)calloc(nmbrinputs(somegate), sizeof(FANIN));
	    }
	  for(k=0; k<nmbrinputs(somegate/*PPOs(newcircuit, j)*/); k++)
	    {
	      faningate(PPOs(newcircuit, j), k) = faningate(somegate, k);
	    }
	  /*FANOUT *outputlist */
	  /* PPOs will not have a fanout list */
	  /* nmbr_inputs */
	  nmbrinputs(PPOs(newcircuit, j)) = nmbrinputs(somegate);
	  /* nmbr_outputs */
	  nmbroutputs(PPOs(newcircuit, j)) = 0/*  nmbroutputs(somegate) */;
	  /* queue_flag */
	  queueflag(PPOs(newcircuit, j)) = queueflag(somegate);
	  /* connected to po */
	  PPOs(newcircuit, j)->connected_to_po = 0/*somegate->connected_to_po*/;
	  /* logic_val */
	  /* Attach PPO to this DFF */
	  somegate->ppo_list.gatepntr = PPOs(newcircuit, j);
	  /* Attach aflist of DFF to PPO */
	  PPOs(newcircuit, j)->aflist = somegate->aflist;
	  j++;/* This should be at the end, nothing should be added after this */
	}
      else
	;
    }

  (* r_circuit) = newcircuit;/* Puts address of CIRCUIT in 'newcircuit' in calling function readckt() */
  /* We can do circuit = newcircuit as well since circuit is global */
  return (seq_circuit);
  
} /* r_parse */

/***************************************************************************
  return the node pointer given a gate number
****************************************************************************/
NODE *get_node_pntr(CIRCUIT *ckt, int gatenmbr)
{ /* get_node_pntr */
  int numpi, numpo;
  numpi = nmbr_pi(ckt);
  numpo = nmbr_po(ckt);
  
  if(gatenmbr <= nmbr_pi(ckt))
    return(PIs(ckt, gatenmbr-1));

  gatenmbr -= nmbr_pi(ckt);
  if(gatenmbr <= nmbr_gates(ckt))
    return(gatesx_dup(ckt, gatenmbr-1));
    //return(ckt->Gate_list_dup[gatenmbr-numpi-1]);
  gatenmbr -= nmbr_gates(ckt);
  if(gatenmbr <= nmbr_po(ckt))
    return(POs(ckt, gatenmbr-1));

  gatenmbr -= nmbr_po(ckt);
  if(gatenmbr <= num_dff)
    return(PPOs(ckt, gatenmbr-1));
  
  fprintf(resultsout, "get_node_pntr: Non-existing gate %d.\n", gatenmbr);
  fatal_error("get_node_pntr", ILLEGAL_CR_GATE, NULL);
  exit (-1);

} /* get_node_pntr */

/****************************************************************************/
/****************************************************************************/
void p_dump_new_netlist (CIRCUIT * ckt, FILE * resultsout)
{ /* p_dump_new_netlist */
  struct node ** gate_array;
  struct node ** pi_array;
  struct node ** po_array;
  int gate_cnt;
  int pi_cnt;
  int po_cnt;
  int i, j;
  struct node * compptr;

  gate_cnt = nmbr_gates (ckt);
  pi_cnt = nmbr_pi (ckt);
  po_cnt = nmbr_po (ckt);
  gate_array = gate_list (ckt);
  pi_array = pi_list (ckt);
  po_array = po_list (ckt);
  fprintf (resultsout, "\nDump of all translated wire connections:\n");

  for (i = 0; i < pi_cnt; ++ i)
    {
      fprintf (resultsout, "PI (PI %d) # inputs %d # outputs %d",
	      pi_array[i]->gate_nmbr,
	      pi_array[i]->nmbr_inputs, pi_array[i]->nmbr_outputs);
      if (pi_array[i]->nmbr_outputs > 0)
	{
	  fprintf (resultsout, " with fanouts:\n");
	  for (j = 0; j < pi_array[i]->nmbr_outputs; ++j)
	    {
	      if ((pi_array[i]->output_list) != NULL)
		{
		  compptr = (struct node *)
		    pi_array[i]->output_list[j].gatepntr;
		  if (compptr == NULL) fprintf (resultsout,
						"Error -- null compptr\n");
		  if (compptr->funct == p_po)
		    fprintf (resultsout, "      (PO %d)\n",
			     compptr->gate_nmbr);
			     /* pi_array[i]->output_list[j].comp_number);*/
		  else
		    fprintf (resultsout, "      (%d)\n",
			    compptr->gate_nmbr); 
			    /* pi_array[i]->output_list[j].comp_number); */
		}
	    }
	}
      else
	fprintf (resultsout, "\n");
    }

  for (i = 0; i < po_cnt; ++ i)
    {
      fprintf (resultsout, "PO (PO %d) # inputs %d # outputs %d ",
              po_array[i]->gate_nmbr,
	      po_array[i]->nmbr_inputs, po_array[i]->nmbr_outputs);
      if (po_array[i]->nmbr_inputs > 0)
	{
	  fprintf (resultsout, "with fanins:\n");
	  for (j = 0; j < po_array[i]->nmbr_inputs; ++j)
	    {
	      if ((po_array[i]->input_list) != NULL)
		{
		  compptr = (struct node *)
		    sourcegate (po_array[i],j);
		  if (compptr == NULL) fprintf (resultsout,
						"Error -- null compptr\n");
		  else
		    {
		      if (compptr->funct == p_po)
			fprintf (resultsout, "      (PO %d)\n",
				compptr->gate_nmbr);
		      else
			fprintf (resultsout, "      (%d)\n",
				compptr->gate_nmbr);
		    }
		}
	      else
		fprintf
		  (resultsout, 
		   "po_array[%d]->input_list should not be null.\n",
		   i);
	    }
	}
      else
	fprintf (resultsout, "\n");
    }

  for (i = 0; i < gate_cnt; ++ i)
    {
      fprintf (resultsout, "COMPONENT  (%d) %s # inputs %d # outputs %d ",
	      gate_array[i]->gate_nmbr,
	      r_trans (gate_array[i]->funct),
	      gate_array[i]->nmbr_inputs, gate_array[i]->nmbr_outputs);
      if (gate_array[i]->nmbr_inputs > 0)
	{
	  fprintf (resultsout, 
		   "with %d fanins:\n", gate_array[i]->nmbr_inputs);
	  for (j = 0; j < gate_array[i]->nmbr_inputs; ++j)
	    {
	      if ((gate_array[i]->input_list) != NULL)
		{
		  compptr = (struct node *)
		    sourcegate (gate_array[i],j);
		  if (compptr == NULL) fprintf (resultsout,
						"Error -- null compptr\n");
		  else
		    {
		      if (compptr->funct == p_pi)
			fprintf (resultsout, "      (PI %d)\n",
				compptr->gate_nmbr);
		      else
			fprintf (resultsout, "      (%d)\n",
				compptr->gate_nmbr);
		    }
		}
	      else
		fprintf
		  (resultsout,
		   "gate_array[%d]->input_list should not be null.\n",
			i);
	    }  
	}
      else
	fprintf (resultsout, "\n");
      if (gate_array[i]->nmbr_outputs > 0)
	{
	  fprintf (resultsout,
		   "  with %d fanouts:\n", gate_array[i]->nmbr_outputs);
	  for (j = 0; j < gate_array[i]->nmbr_outputs; ++j)
	    {
	      if ((gate_array[i]->output_list) != NULL)
		{
		  compptr = (struct node *)
		    gate_array[i]->output_list[j].gatepntr;
		  if (compptr == NULL) fprintf (resultsout,
						"Error -- null compptr\n");
		  if (compptr->funct == p_po)
		    fprintf (resultsout, "      (PO %d)\n",
			    compptr->gate_nmbr);
			    /* gate_array[i]->output_list[j].comp_number); */
		  else
		    fprintf (resultsout, "      (%d)\n",
			    compptr->gate_nmbr);
			    /* gate_array[i]->output_list[j].comp_number); */
		}
	    }
	}
      else
	fprintf (resultsout, "\n");
    }
  fflush (resultsout);
  return;

} /* p_dump_new_netlist */

/* */
void duplicate_gate_pntrs(CIRCUIT *ckt, int gatecount)
{/* duplicate_gate_pntrs */
  int i;

  ckt->Gate_list_dup = (NODE **)calloc(gatecount, sizeof(NODE *));
  if(ckt->Gate_list_dup == NULL)
    {
      printf("Memory for Duplicate Gate Pointers NOT Allocated!\n");
      exit(-1);
    }
  else
    {
      for(i=0; i<gatecount; i++)
	{
	  gatesx_dup(ckt, i) = gatesx(ckt, i);
	}
      printf("Gate List Pointers Copied!\n");
    }
}/* duplicate_gate_pntrs */
