/************************************************************************
     serialqueue.c
*************************************************************************/
#include "all.h"

/*************************************************************************/
             /* s_enqueue() is called whenever we queue an event.
		Function: Queue the event at the tail of the current 
		          event queue.
			  Don't queue if the gate is already queued 
		Side effects: Modifies the tail pointer of the queue
		              to point to the new free location.
		              Prints error message if queue overflows and
			      exits with error code -26.
	      */
/*************************************************************************/
void s_enqueue(NODE *gate)
     /* The gate number for which the changes are to be made */
{ /* s_enqueue */

  if(queueflag(gate) == FALSE)
                             /* queue only if the gate is not already
				queued */
    {
      /* set queueflag to not queue this gate again */
      queueflag(gate) = TRUE;

      /* Queue information */
      s_queue[s_queue_tail] = gate;
      
      if(++s_queue_tail >= QUEUE_SIZE) /* If the tail index increases beyond
					   the size of the event queue,
					   roll back */
	s_queue_tail = 0;
#ifdef RUDEBUG
      /* check if queue overflowed if desired.  This check is not required
	 normally because the queue never grows beyond gatecount because of
	 the queue flag stuff */
      if(s_queue_head == s_queue_tail)
	{
	  if(VERBOSE == TRUE)
	    fprintf(resultsout, "enqueue: Queue overflow.\n");
	  fatal_error("enqueue", QUEUE_OVERFLOW, NULL);
	}
#endif
    }
} /* s_enqueue */

/*************************************************************************/
              /* s_dequeue() is called whenever we dequeue next event from
		 an event queue.
		 Function: Returns information pertaining to the gate 
		           for which changes are to be affected.  Specifically,
			   gate number, its input number, the new value, and
			   the information whether the change is at the INPUT
			   or at the OUPUT are returned.
		 Side effects: Modifies the value of the head pointer to the
		               event queue.
	       */
/*************************************************************************/
NODE *s_dequeue(void)
{ /* s_dequeue */
  NODE *gate = s_queue[s_queue_head];

  if(++s_queue_head >= QUEUE_SIZE)    /* If the head index of the event
				       queue increases beyond the queue
				       size, then roll back */
    s_queue_head = 0;

  /* Nullify the queueflag to show that this gate is no longer
     queued */
  if (gate != NULL) queueflag(gate) = FALSE;
  return(gate);
} /* s_dequeue */


/*************************************************************************/
             /* queueinit() is called at the beginning of the program 
		execution to initialize the event queues used by the 
		simulation routines.
		Function: Dynamically allocates memory space for the 
		          queue array.  Initializes the head and 
			  tail indicies of the queue to zero.
	      */
/*************************************************************************/
void s_queueinit(int size)
{ /* s_queueinit */
  
  s_queue = (NODE **)calloc(size, sizeof(NODE *));
  if(s_queue == NULL)
    {
      fatal_error("s_queueinit", MEMORY_ALLOC_FAILURE, NULL);
    }
  s_queue_head = 0;
  s_queue_tail = 0;
} /* s_queueinit */

void free_s_queue(void)
{ /* free_s_queue */
  free(s_queue);
} /* free_s_queue */

/****************************************************************************
        clear_s_queue() clears the given queue.
	Inputs: cur_queue --> Pointer to the queue that is to be cleared
	Function: Initialize head and tail pointers both to zero.
	Return Value: none.
****************************************************************************/
void clear_s_queue(void)
{ /* clear_s_queue */
  s_queue_head = 0;
  s_queue_tail = 0;
} /* clear_s_queue */

/****************************************************************************
       s_queue_not_empty() checks if the queue is empty.
	Function: Check if the head and tail indicies are equal.
	          If equal then the queue is empty.
	Return value: 0 if empty, non-zero if not empty
****************************************************************************/
int s_queue_not_empty(void)
{ /* s_queue_not_empty */
  return(s_queue_head != s_queue_tail);
} /* s_queue_not_empty */


